/*
 * Copyright (c) 2019, Texas Instruments Incorporated
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * *  Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * *  Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * *  Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

const COMPATIBLE_EVMID_REGEX = /INA229EVM/;
const LATEST_FIRMWARE_VERSION = 20200323160603;


// Global EVM object
var EVM = {
    id: {
        name: "",
        date: "",
        time: "",
        version: ""
    },
    preamble: { }
};

var Terminal = false;

/*
 * gc global variable provides access to GUI Composer infrastructure components and project information.
 * For more information, please see the Working with Javascript guide in the online help.
 */
var gc = gc || {};
gc.services = gc.services || {};


/* 
 * To see the console logs logged by this file, in the console view, enter the following:
 * gc.console.setCookie("pambCustomCodec.js", <value>) 
 * Where value: 0 = remove the cookie, 1 = only errors, 2 = errors and warnings, 3 = errors,warnings,info, 4 = errors,warnings,info,logs, 5=debug+all others
 * For more help on how to work with gc.console, please see the help text that is sent to the
 * console when the app starts.
 */
//if (!gc.designer) { gc.console.setCookie("pambCustomCodec.js", 5); }      // NOTE: Changing to setLevel() in GC3



document.addEventListener('gc-databind-ready', function() {

    gc.console.debug("pambCustomCodec.js", "The 'gc-databind-ready' event fired, building CustomCodec object.");

    /**
     * This code implements the AbstractMessageBasedCodec interface.
     *
     * @constructor
     * @implements gc.databind.AbstractMessageBasedCodec
     */
    CustomCodec = function() {};
    
    // IMPORTANT: Do NOT add any objects to the CustomCodec's protoype before calling the
    // gc.databind.AbstractMessageBasedCodec(), as they will get overwritten!
    
    //////////////////////// AbstractMessageBasedCodec APIs ////////////////////////
    
    // See additionals APIs and object provided by 'ti-core-databind\src\AbstractMessageBasedCodec.js'
    CustomCodec.prototype = new gc.databind.AbstractMessageBasedCodec("PAMB");
    // Objects: name, commandQueue, disconnect()
    
    // Overwriting the default 'connect' method in AbstractMessageBasedCodec...
    CustomCodec.prototype.connect = function(controllerConfig) {
        
        //
        // IMPORTANT: This method is called by the transport layer when attempting to connect to the serial COM port.
        // The "controllerConfig" object comes from the 'system.json' file.
        //
        
        // Update the function scope for the AbstractMessageBasedCodec connect method
        gc.databind.AbstractMessageBasedCodec.prototype.connect.call(this);
        
        // Return a promise that will resolve after connection is complete
        return Q.Promise(async function(resolve, reject) {
            
            // Set initial firmware state?
            this.setState("idle");                  // TODO: Make sure the firmware resets to idle state after connecting
            
            // Set connected flag
            this.serialDeviceConnected = true;
            
            // Get EVM ID (timeout if ID is not returned after 5 seconds)...
            let evmId = await this.sendCommand(EVM_ID_COMMAND, 5000);
            
            // TODO: Read EEPROM
            
            // Abort execution if EVM ID is unknown...
            if (!evmId) {
                gc.console.error("pambCustomCodec.js", `Could not resolve EVM ID!`);
                reject("Communication with EVM timed out");
                return;
            }
            
            // Test EVM name versus regex expression to see if it is compatible
            let isCompatible = COMPATIBLE_EVMID_REGEX.test(evmId.name);
            gc.console.debug("pambCustomCodec.js", `isCompatible = ${isCompatible}`);
            if (!isCompatible) {
                // Reject connect promise -> This will call the disconnect() method
                gc.console.debug("pambCustomCodec.js", "Rejecting 'connect' promise");
                reject("This application is not compatable with the " + EVM.id.name + ". Please \
                try a different EVM or GUI Composer application!");
                return;
            }
            
            // Check if an EVM firmware update is out-of-date
            if (evmId.version < LATEST_FIRMWARE_VERSION) {
                
                // Release the USB interface
                reject("Updating the EVM firmware...");
                
                // Send Invoke BSL command
                this.encoder(EVM_INVOKE_BSL_COMMAND);
                
                alert('Firmware out of date. Will now update firmware.')
                
            } else {
                
                // TODO: Should we wait for the bulk device to connect?
                
                // Resolve connect promise
                gc.console.debug("pambCustomCodec.js", "Resolving 'connect' promise");
                resolve();
            }
            
        }.bind(this));  // Since this is a type of callback function, we must bind 'this' to the CustomCodec object.
    };
    
    // Overwriting the default 'disconnect' method in AbstractMessageBasedCodec...
    CustomCodec.prototype.disconnect = function() {
        
        // Call AbstractMessageBasedCodec's disconnect method
        gc.databind.AbstractMessageBasedCodec.prototype.disconnect.call(this);
        
        // Set connection flag
        this.serialDeviceConnected = false;
        
        // Clear all pending commands and deferred writeValue and readValue promises
        this.pendingSendCmds = [];
        //this.commandQueue.clearAll();   // This is handled by the AbstractMessageBasedCodec's disconnect method 
        
        // Call close function()
        closeBulkDevice(false); // 'false' indicates that this is not in response to a detach event
        
        // NOTE: Remove all event listeners you added in the connect function here,
        // unless they were only called once and removed from within the callback function.
        // See reference: https://stackoverflow.com/questions/23805878/why-call-removeeventlistener-inside-addeventlistener-callback
    };
    
    // Overwriting the default 'isStillConnected' method in AbstractMessageBasedCodec...
    // This function is periodically called by the internal state machine to check if hardware is connected.
    CustomCodec.prototype.isStillConnected = function() {
        
        // (OPTIONAL) Ping the hardware and check for a response
        
        // Resolve promise based on our internal state flag
        if (this.serialDeviceConnected)     { return Q.resolve(true); }
        else                                { return Q.reject(false); }
    };
    
    
    
    /////////////////////////// IPollingPacketCodec APIs ///////////////////////////
    
    // NOTE: AbstractMessageBasedCodec also implements the IPollingPacketCodec
    // CustomCodec.prototype = new gc.databind.IPollingPacketCodec();
    
    /**
     * Encodes data into a packet for sending to the target device.
     *
     * @param target - function to call to pass on the encoded data towards the target
     * @param jsonObj - The data object from the GUI that is to be encoded into a packet for sending.
     */
    CustomCodec.prototype.encode = function(target, cmdString) {
        
        // Do not send commands while EVM is busy.
        if (this.evmState === "busy") {
            
            // Enque command
            this.enqueueCmd(cmdString);
            return;
            
        } else if ((this.evmState === "collecting") && (cmdString !== "collectstop")) {
            
            // Command not allowed while collecting data
            return;
        }
        
        // Send command...
        this.strTxBuffer = cmdString; // We will use this buffer for recalling the last command sent to the device
        if (this.strTxBuffer.length > 0) {
            
            // Update EVM state to busy
            this.setState("busy");
            
            // Debugging...
            let msg = "Sending command '" + this.strTxBuffer + "'";
            if (Terminal) { Terminal.writeTextArea("\n" + msg + "\n"); }
            
            // Log message (NOTE: Separator is used to make the log easier to read)
            gc.console.info("pambCustomCodec.js", "------------------------------------------");
            gc.console.info("pambCustomCodec.js", msg);
            
            try {
                target(this.strTxBuffer);
            } catch (ex) {
                gc.console.error("pambCustomCodec.js", "CustomCodec.encode: exception=" + ex);
            }
        }
        
        // Set 3 second timeout in case firmware does not respond to command
        // NOTE: The timer will be cleared when the firmware acknowledges the command.
        // this.timerID = setTimeout(function() {
        //     this.serialDeviceConnected = false;
        //     alert("Timeout: Firmware failed to acknowledge '" + this.strTxBuffer + "' command within 3 seconds!");
        // }.bind(this), 3000);
    };
    
    
    /**
     * Decodes packets from the target device into data objects.  One object for each packet of data.
     * Unless you have chained your custom codec with the CR codec, the packet data is not framed,
     * so this method is responsible for decoding partial packets as well as multiple packets.  Partial packets
     * should not be returned, but deferred until the next decode method call with more raw data to process.  As a
     * result, this method may return 0 or more packets of data on each invocation.
     *
     * @param target -  function to call to pass on a JSON object containing the decoded values towards the GUI
     * @param data  - The raw data received that needs to be decoded from packets into objects.
     * @return - true if serialDeviceConnected (i.e. valid data received), false if not connected
     */
    CustomCodec.prototype.decode = function(target, data) {
        
        // Convert data to ASCII
        var strRxData = data.map(function(value) {
            return String.fromCharCode(value);
        }).join("");
        
        // Add RX data to strRxBuffer and look for a newline character...
        this.strRxBuffer = this.strRxBuffer + strRxData;
        while (this.strRxBuffer.indexOf('\n') !== -1) {
            
            // Parse JSON message from strRxBuffer (message must always end with newline character!)
            var strMessage = this.strRxBuffer.substring(0, this.strRxBuffer.indexOf('\n') + 1);
            gc.console.info("pambCustomCodec.js", "Received response " + strMessage);
                
            // Remove JSON message from strRxBuffer
            this.strRxBuffer = this.strRxBuffer.substring(strMessage.length);
                
            // Check for a message to parse
            if (!strMessage) { break; }
                
            // Update Terminal text area
            if (Terminal) { Terminal.writeTextArea(strMessage); }
                
            // Try to parse message as a JSON object
            try {
                
                // Parse JSON string to object
                var jsonObj = JSON.parse(strMessage.trim());
                
                //
                // Look for predefined JSON properties to decide what action(s) to take...
                //
                
                // EVM firmware is acknowledging a recieved command...
                if (jsonObj.hasOwnProperty("acknowledge")) {
                    
                    // Clear firmware timeout (if we set one)
                    clearTimeout(this.timerID);
                    
                    // Get acknowledged command string
                    let cmd_ack = jsonObj.acknowledge;
                    
                    // TODO: Check if we got an acknowledge to the 'bsl' command and show message to user...
                    // See https://bitbucket.itg.ti.com/projects/SOL_GCS/repos/saradc/commits/d7dffbf99b18f06136c76aae6622dfc7f3501825#app/pambCustomCodec.js
                    
                    // Update Terminal
                    if (Terminal) { Terminal.acknowledge(cmd_ack); }
                    
                    // Check if this response does not match the last command sent to the EVM...
                    if (cmd_ack !== this.strTxBuffer) {
                        gc.console.error("pambCustomCodec.js", "The EVM's acknowledged command ('" + cmd_ack + "') does not match the last command sent ('" + this.strTxBuffer + "')");
                    }
                    
                    // NOTE: In the case where writeValue() is waiting on a promise, you may want to resolve that promise here:
                    // this.resolveCmdPromise(/^wreg/, undefined, false);
                    
                }
                
                // Log firmware message to console...
                if (jsonObj.hasOwnProperty("console")) {
                    
                    // Get message
                    let msg = jsonObj.console;
                    
                    // Check for message prefix of "Error", "Warn", or "Info" to direct log output type
                    let msgType;
                    if (msg.substr(0, 5).toLowerCase() === "error") {
                        msgType = "error";
                        gc.console.error("pambCustomCodec.js", msg);
                        
                    } else if (msg.substr(0, 4).toLowerCase() === "warn") {
                        msgType = "warn";
                        gc.console.warn("pambCustomCodec.js", msg);
                        
                    } else {
                        msgType = "info";
                        gc.console.log("pambCustomCodec.js", msg);
                    }
                    
                    // Output message to application's eventlog widget
                    if (this.eventLog) { this.eventLog.log(msgType, msg); }
                }
                
                // Firmware is reporting an error...
                if (jsonObj.hasOwnProperty("error")) {
                    // TODO: build list of possible error numbers
                }
                
                // (OPTIONAL) EVM device mode has changed...
                if (jsonObj.hasOwnProperty("device_modes")) {
                    // Extract mode information and store it for reference...
                }
                
                // EVM ID
                if (jsonObj.hasOwnProperty("id")) {
                    
                    // Update the EVM object
                    EVM.id = jsonObj.id;
                    EVM.id.version = this.firmwareVersionNumber(jsonObj.id);
                    
                    // Print info to event log
                    if (this.eventLog) { this.eventLog.log("EVM identified as '" + EVM.id.name + "'. Firmware version = '" +  EVM.id.version + "'."); }
                    
                    // Print info to gc console
                    gc.console.log("pambCustomCodec.js", "EVM identified as '" + EVM.id.name + "'. Firmware version = '" +  EVM.id.version + "' (" + EVM.id.date + " " + EVM.id.time + ")");
                    
                    // Resolve promise (if any)
                    this.resolveCmdPromise(/^id/, EVM.id, false);
                    
                }
                
                // Information about the binary data formatting
                // IMPORTANT: THIS WILL NEED TO BE UPDATED FOR EACH EVM GUI!
                // TODO: Move to a collect function
                if (jsonObj.hasOwnProperty("preamble")) {
                    
                    //
                    // Build FrameDefinitions array from current register map settings
                    //
                    
                    // Get number of channels...
                    let idVal = this.getRegisterValue(0x00);   // 0x00 = address of ID reigster
                    channelCount = (idVal & 0x0F00) >> 8;
                    
                    // Get data byte length
                    let modeVal = this.getRegisterValue(0x02);   // 0x02 = address of MODE reigster
                    let wlengthVal = (modeVal & 0x0300) >> 8;
                    bytesPerWord = (wlengthVal < 3) ? wlengthVal+2 : 4; // Convert field value to number of bytes, per device datasheet
                    
                    //
                    // FRAME 0
                    //
                    FrameDefinitions[0x00] = [];    // NOTE: When building a 2D array, we need to initialize each frame separately.
                    
                    // Frame ID (required) 
                    FrameDefinitions[0x00].push(new CreateChannel("frameId", 1, false, null));
                    
                    // STATUS
                    FrameDefinitions[0x00].push(new CreateChannel("status", bytesPerWord, false, statusArray));
                    
                    // CHANNELS 
                    for (let i = 0 ; i < channelCount; i++) {
                        // NOTE: We create a channel data placeholder for all channels, even if they are disabled
                        channelData[i] = [];
                        FrameDefinitions[0x00].push(new CreateChannel("channel" + i, bytesPerWord, true, channelData[i]));
                    }
                    
                    // CRC
                    FrameDefinitions[0x00].push(new CreateChannel("crc", bytesPerWord, false, crcArray));
                }
                
                // Update register data
                if (jsonObj.hasOwnProperty("register")) {
                    
                    // Get register read value
                    let readValue = jsonObj.register.value;
                    
                    // Check if we are resolving a promise...
                    let resolved = this.resolveCmdPromise(/^rreg/, +readValue, false);
                    if (!resolved) {
                        
                        // Update the register model directly
                        this.setRegisterValue(jsonObj.register.address, readValue);
                        
                    }
                }
                
                // EVM state (tells us if EVM is 'idle' and ready to recieve commands)
                if (jsonObj.hasOwnProperty("evm_state")) {
                    this.setState(jsonObj.evm_state);
                }
                
                // Send the received and decoded javascript object to the application where you can use it to bind to widget properties
                target(jsonObj);
                
            } catch (ex) {
                gc.console.error("pambCustomCodec.js", "decode (from target): strMessage='" + strMessage + "', exception=" + ex);
            }
        }
        
        // Return true to set the 'Hardware Connected' status in the status bar
        return this.serialDeviceConnected;
    };
    
    
    
    /////////////////////////
    // Register model APIs //
    /////////////////////////
    
    /*
     * Confluence: https://confluence.itg.ti.com/pages/viewpage.action?pageId=173550282
     * For an example of a codec that accesses symbols using readValue/writeValue, please see components/ti-widget-registers/demo/MotorwareDemo_F28069M/customCodec_0.js
     */
    
    // Called by register model to read a device register...
    CustomCodec.prototype.readValue = function(regInfo) {
        
        // (OPTIONAL) Check for valid register address and reject promise if invalid
        //if (!findRegister(regInfo.addr)) {
        //    gc.console.warn("pambCustomCodec.js", "REJECTING readValue : " + regInfo.addr);
        //    return Q.reject("Invalid register address");
        //}
        
        // Add the pending command to the internal commandQueue APIs provided by 
        // ti-core-databind\src\AbstractMessageBasedCodec.js
        let addressString = (regInfo.addr).toString(16);
        let cmd = READ_REGISTER_COMMAND + ' ' + addressString;
        let promise = this.addCommand(cmd, this.pendingCmdSeqNum++);
        
        // Debugging...
        gc.console.log("pambCustomCodec.js", `readValue(): Register model requesting data from register 0x${addressString}`);
        
        // Send read register command
        this.encoder(cmd);
        
        return promise;
    };
    
    // Called by register model to write register data...
    CustomCodec.prototype.writeValue = function(regInfo, value) {
        
        // Build the write register command string
        let addressString = (regInfo.addr).toString(16);
        let valueString = value.toString(16);
        let cmd = WRITE_REGISTER_COMMAND + ' ' + addressString + ' ' + valueString;
        
        // Debugging...
        gc.console.log("pambCustomCodec.js", `writeValue(): Register model requesting write data 0x${valueString} to register 0x${addressString}`);
        
        // Send write register command to queue the command locally
        this.encoder(cmd);
        
        // Return a completed promise (in this case with undefined value since writeValue does not look at the result, it just waits for the promise to complete.
        return Q();
        
        // NOTE: There is no need to delay the completion of the writeValue promise,
        // but if for some reason you need to, then here is an alternate implementation.
        // Add the pending command to the internal commandQueue APIs provided by 
        // ti-core-databind\src\AbstractMessageBasedCodec.js and get a promise
        // let promise = this.addCommand(cmd, this.pendingCmdSeqNum++);
        // gc.console.log("pambCustomCodec.js", "writeValue: this.commandQueue.nextCmd.command = " + this.commandQueue.first.next.command);
        // return promise;
    };
    
    // Lookup function to get the fully qualified register name from the register address.
    CustomCodec.prototype.findRegister = function(regAddrs) {
        
        // Coerce argument to number
        let address = +regAddrs;
        if (isNaN(address)) {
            gc.console.error("pambCustomCodec.js", `findRegister(): requires a numeric argument, recieved ${address}`);
            return;
        }
        
        // If no cached info, then we cannot perform the lookup
        if (!this.cachedRegisters) {
            gc.console.error("pambCustomCodec.js", "findRegister(): 'cachedRegisters' is not defined. Cannot perform lookup!");
            return;
        }
        
        // If no register definitions (i.e. addresses) are skipped in the JSON register map file,
        // then we can simply index the array...
        let result = this.cachedRegisters[address];
        if (result.addr === address)  { return result; }
        
        // Else...loop through register definitions looking for the matching register address...
        const matchingRegisters = this.cachedRegisters.filter(r => r.addr === address);
        const numMatches = matchingRegisters.length;
        if (numMatches !== 1) {
            gc.console.error("pambCustomCodec.js", `findRegister(): found ${numMatches} registers with address = ${address}`);
        } else {
            return matchingRegisters[0];
        }
    };
    
    // Lookup function to get the current register value from a specified register address in the register model.
    CustomCodec.prototype.getRegisterValue = function(regAddrs) {
        
        // Get register object which contains register info
        let regObj = this.findRegister(regAddrs);
                    
        // Get register value from register model...
        if (regObj) {
            let regBind = gc.databind.registry.getBinding(this.registerModel.id + '.' + regObj.fullyQualifiedName);
            return regBind.getValue();
        } 
    };
    
    // Lookup function to set a register value from a specified register address in the register model.
    CustomCodec.prototype.setRegisterValue = function(regAddrs, value) {
        
        // Get register object which contains register info
        let regObj = this.findRegister(regAddrs);
        
        // Update the register model with the new register data...
        if (regObj) {
            let regBind = gc.databind.registry.getBinding(this.registerModel.id + '.' + regObj.fullyQualifiedName);
            regBind.updateValue(value);
        }
    };
    
    
    //////////////////////// Custom Properties ////////////////////////
    
    // TODO: Define additional fimrware commands here...
    const WRITE_REGISTER_COMMAND = 'wreg';
    const READ_REGISTER_COMMAND = 'rreg';
    const EVM_ID_COMMAND = 'id';
    const EVM_INVOKE_BSL_COMMAND = 'bsl';
    
    // RX and TX buffers
    CustomCodec.prototype.strRxBuffer = "";
    CustomCodec.prototype.strTxBuffer = "";
    
    // Command queueing
    CustomCodec.prototype.evmState = "";
    CustomCodec.prototype.pendingSendCmds = [];
    CustomCodec.prototype.pendingCmdSeqNum = 0;
    CustomCodec.prototype.cmdSeqNum = 0;
    
    // Internal state flags
    CustomCodec.prototype.serialDeviceConnected = false;
    CustomCodec.prototype.bulkDeviceConnected = false;
    
    // References to important objects
    CustomCodec.prototype.registerModel     = undefined;    // Reference to register model object
    CustomCodec.prototype.cachedRegisters   = undefined;    // Placeholder for saving register model information
    CustomCodec.prototype.eventLog          = undefined;    // Reference to eventLog object
    CustomCodec.prototype.timerID           = undefined;    // Timer ID for firmware timeout
    
    // Save reference to register model, and cache the register data
    this.registerModel = document.querySelector('ti-model-register');
    if (this.registerModel) {
        
        // Cache register block information
        let deviceIndex = 0;
        this.registerModel.getRegisterViewInformation(deviceIndex).then(function(deviceObj) {
            
            // Flatten
            this.cachedRegisters = [];
            deviceObj.regblocks.forEach(block => block.registers.forEach(reg => { this.cachedRegisters.push(reg); }));
            
            // Sort
            this.cachedRegisters = this.cachedRegisters.sort((registerA, registerB) => {
                if      (registerA.addr > registerB.addr) { return 1; }
                else if (registerB.addr > registerA.addr) { return -1; }
                else {
                    gc.console.error("pambCustomCodec.js", `Cannot have two registers with the same address (${registerA.name} and ${registerB.name}, address = ${registerA.addr})!`);
                }
            });
            
        }.bind(this)); // NOTE: Callback function must be called from the current scope
        
    }
    
    // Save reference to eventlog
    this.eventLog = document.querySelector('ti-widget-eventlog-view');
    
    
    ///////////////////////// CustomCodec Helper functions /////////////////////////
    
    // Sends a command to the EVM firmware (with an optional timeout) and returns a promise with the response
    CustomCodec.prototype.sendCommand = function(cmdString, msTimeout) {
        
        // Coerce
        msTimeout = +msTimeout || 0;
        
        // Debugging...
        let msg = `sendCommand('${cmdString}', ${msTimeout})`;
        gc.console.debug("pambCustomCodec.js", msg);
        
        // Add the pending command to the internal commandQueue
        let cmdPromise = this.addCommand(cmdString, this.pendingCmdSeqNum++);
        let returnPromise = cmdPromise;
        
        if (msTimeout) {
            let racePromise = MyUtil.promiseTimeout(msTimeout, cmdPromise)
            .then(result => {
                return result;  // cmdPromise completed
            })
            .catch(error => {
                // The cmdPromise was rejected or the timeout occurred (likely the latter)
                msg = `'${EVM_ID_COMMAND}' command was rejected with reason: '${error}'`;
                gc.console.warning("pambCustomCodec.js", msg);
                
                // TODO: Try to remove command from queue, re-establish communication with evm (send a stop command), or disconnect
                // NOTE: The current implementation will allow the command to complete even after the timeout has expired.
                //debugger;
                return;
            });
            
            // Update the return promise pointer to racePromise (since this is the promise we want to wait for)
            returnPromise = racePromise;
        }
        
        // Read ID of the EVM and make sure it matches
        this.encoder(cmdString);
        
        return returnPromise;
    };
    
    // Update EVM state variable
    CustomCodec.prototype.setState = function(state) {
        
        // Check if state is valid...
        const re = new RegExp('^idle$|^busy$|^collecting$');
        const isValid = re.test(state);
        if (!isValid) {
            gc.console.error("pambCustomCodec.js", `Invalid state argument in setState(): '${state}'`);
            return;
        }
        
        // Update state variables
        let prevState = this.evmState;
        this.evmState = state;
        
        // Update state indicator
        if (Terminal) { Terminal.setState(state); }
        
        // Handle state transistions...
        switch (state) {
            
            case "idle":
                if (prevState == "collecting") {
                    // Analyze new data
                    channelData[0];     // TODO: Implement
                }
                
                // Dequeue next command (if any)...
                // NOTE: Make sure to update the internal state variable before
                // calling this function, or else the dequeuing method will fail!
                this.dequeueCmd();
                break;
                
            case "busy":
                break;
                
            case "collecting":
                
                // Clear old data
                Parser.clear(FrameDefinitions);
                break;
        }
    };
    
    // Enqueue command
    CustomCodec.prototype.enqueueCmd = function(cmdString) {
        
        // Enque command
        this.pendingSendCmds.push({
            cmd: cmdString,
            //promise: this.addCommand(cmd, this.pendingCmdSeqNum++) // NOTE: We may not want to create a promise with each command...
        });
        gc.console.debug("pambCustomCodec.js", "Enquing command '" + cmdString + "' (" + this.pendingSendCmds.length + " cmds pending)");
    };
    
    // Dequeue command
    CustomCodec.prototype.dequeueCmd = function() {
        
        // Check if we have any pending commands to dequeue
        if (this.pendingSendCmds.length > 0) {
            let nextCmd = this.pendingSendCmds.shift();
            gc.console.debug("pambCustomCodec.js", "Dequing command '" + nextCmd.cmd + "' (" + this.pendingSendCmds.length + " cmds pending)");
            
            this.encoder(nextCmd.cmd);
        }
    };
    
    // Resolve command promise
    CustomCodec.prototype.resolveCmdPromise = function(regex, response, error) {
        
        // Check if we are resolving a promise...
        let pendingCmd = this.commandQueue.first.next;
        if (!pendingCmd || !pendingCmd.command) { return false; }
        
        if (regex.test(pendingCmd.command)) {
            
            let msg = "Resolving '" + pendingCmd.command + "' command promise [seqNo: " + pendingCmd.seqNo + ", isError: " + error + "] with " + JSON.stringify(response);
            gc.console.log("pambCustomCodec.js", msg);
            
            // Resolve promise based on match result
            this.addResponse(response, pendingCmd.command, pendingCmd.seqNo, error);
            this.cmdSeqNum++;
            
            if (!pendingCmd.next) {
                gc.console.debug("pambCustomCodec.js", "The commandQueue is now empty");
            }
            
            return true;    // Indicate to caller that we resolved a promise
        }
        else { return false; }  // Indicate to caller that we did not resolve a promise
    }
    
    // Build firmware version number from compiler's date/time string
    CustomCodec.prototype.firmwareVersionNumber = function(idObj) {
        
        let date = idObj.date.split(" "); // Ex: 'Feb','','7','2019'
        date = date.filter(entry => entry.trim() !== ''); // Remove any null entries that occur on single digit days
        
        let time = idObj.time.split(":"); // Ex: '16','48','59'
        
        // REFERENCE: http://infocenter.arm.com/help/index.jsp?topic=/com.arm.doc.faqs/ka9555.html
        const month_names = ["Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"];
        
        // Get index of matching month abbreviaiton
        let month_index = 1 + month_names.findIndex(function(element) {
            return (element === date[0]);
        });
        
        // Convert values to strings, with length 2 (except for year)
        let year = date[2];
        let month = MyUtil.pad(month_index, 2); //= (month_index < 10) ? ('0' + month_index) : month_index;
        let day = (date[1].length < 2) ? ('0' + date[1]) : date[1];
        let hour = (time[0].length < 2) ? ('0' + time[0]) : time[0];
        let min = (time[1].length < 2) ? ('0' + time[1]) : time[1];
        let sec = (time[2].length < 2) ? ('0' + time[2]) : time[2];
        
        // Concatenate strings
        let versionString = ("" + year + month + day + hour + min + sec);
        return versionString;
    };
    
    CustomCodec.prototype.collectData = function() {
        // TODO: Determine how to stream or request a specific number of samples..
    };
    
    
    ///////////////////////////// Register CustomCodec /////////////////////////////
    
    // registerCustomCodec has to be called right at startup in order for the codec to be recognized and used
    gc.console.log("pambCustomCodec.js", "CustomCodec: about to register...");
    
    // You can 'chain' codecs together to make it easier to work with the data
    // coming from the device or going to the device.
    // The following codecs are provided:
    //   CR  - use this if the target data uses /n as a delimiter for framing.
    //         The CR codec will take care of buffering the incoming data so that
    //         your codec will only receive complete 'ready to use' packets
    //  If your target uses \n delimited strings, set baseCodecs = "CR"
    //  If your target uses some other format for data (e.g. binary data or custom strings) leave baseCodecs = "undefined"
    var baseCodecs;
    gc.databind.registerCustomCodec("custom", CustomCodec, baseCodecs);

    gc.console.debug("pambCustomCodec.js", "CustomCodec object registered.");
    
}); // end 'gc-databind-ready' eventlistener function



///////////////////////// Binary data parsing classes //////////////////////////

// TODO: Move these classes to usb.js...

// Array buffer methods for converting binary to integer types
var ArrayBufferCursor = (function() {
    "use strict";
    
    // REFERENCE: https://riptutorial.com/javascript/example/32392/iterating-through-an-arraybuffer
    
    // Constructor
    var ArrayBufferCursor = function(arrayBuffer) {
        this.dataview = new DataView(arrayBuffer, 0);
        this.size = arrayBuffer.byteLength;
        this.index = 0;
        this.isLittleEndian = false;
    };
    
    ArrayBufferCursor.prototype.next = function(type) {
        let result;
        let unsignedINT24;
        switch(type) {
            
            // NOTE: For 8-bit data types, the array buffer index is being incremented inline...
            
            case 'Int8':
                result = this.dataview.getInt8(this.index++, this.isLittleEndian);
                return result;
            
            case 'Uint8':
                result = this.dataview.getUint8(this.index++, this.isLittleEndian);
                return result;
                
            case 'Int16':
                result = this.dataview.getInt16(this.index, this.isLittleEndian);
                this.index += 2;
                return result;
                
            case 'Uint16':
                result = this.dataview.getUint16(this.index, this.isLittleEndian);
                this.index += 2;
                return result;
                
            case 'Int24':
                unsignedINT24 = this.dataview.getUint8(this.index++, this.isLittleEndian);
                unsignedINT24 = (unsignedINT24 << 8) + this.dataview.getUint8(this.index++, this.isLittleEndian);
                unsignedINT24 = (unsignedINT24 << 8) + this.dataview.getUint8(this.index++, this.isLittleEndian);
                return ((unsignedINT24 & 0x800000) ? (unsignedINT24 - Math.pow(2, 24)) : unsignedINT24);
                
            case 'Uint24':
                unsignedINT24 = this.dataview.getUint8(this.index++, this.isLittleEndian);
                unsignedINT24 = (unsignedINT24 << 8) + this.dataview.getUint8(this.index++, this.isLittleEndian);
                unsignedINT24 = (unsignedINT24 << 8) + this.dataview.getUint8(this.index++, this.isLittleEndian);
                return unsignedINT24;
                
            case 'Int32':
                result = this.dataview.getInt32(this.index, this.isLittleEndian);
                this.index += 4;
                return result;
                
            case 'Uint32':
                result = this.dataview.getUint32(this.index, this.isLittleEndian);
                this.index += 4;
                return result;
                
            // UNSUPPORTED TYPES...
            // case 'Float32':
            //     result = this.dataview.getFloat32(this.index, this.isLittleEndian);
            //     this.index += 4;
            //     return result;
            
            // case 'Float64':
            //     result = this.dataview.getFloat64(this.index, this.isLittleEndian);
            //     this.index += 8;
            //     return result;
            
            default:
                throw new Error("Unknown datatype");
        }
    };
    
    ArrayBufferCursor.prototype.hasNext = function(typeString) {
        if (typeString) {
            
            // Get number of bytes from type string (e.g. 'Uint24' => 24 / 8 = 3)
            var numberBytes = (typeString.match(/\d+/)) / 8;
            
            // Check if there are enough bytes to extract this data type
            return (this.size - this.index) >= numberBytes;
            
        } else {
            
            // Check if there is at least 1 byte in arraybuffer
            return this.index < this.size;
        }
    };

    return ArrayBufferCursor;
})();

// Binary data parser for sorting data
var BinaryDataParser = (function() {
    "use strict";
    
    // Constructor
    var BinaryDataParser = function() {
        this.frameId = 0;
        this.channelCounter = 0;
    };
    
    BinaryDataParser.prototype.getCursor = function(data) {
        var ArrayObj = new Uint8Array(data);
        return new ArrayBufferCursor(ArrayObj.buffer);
    };
    
    BinaryDataParser.prototype.parse = function(binaryDataArray, dataFrames) {
        
        // Get cursor
        var cursor = this.getCursor(binaryDataArray);
        
        // Get current frame
        var currentFrame = dataFrames[this.frameId];
        if (!currentFrame) {
            gc.console.error("pambCustomCodec.js", "Data parsing misalignment: Invalid frame id: " + this.frameId + "!");
        }
        
        // Get current channel
        var channel = currentFrame[this.channelCounter];
        if (!channel) {
            gc.console.error("pambCustomCodec.js", "Data parsing error: Invalid channel: " + this.channelCounter + "!");
        }
        
        // Parse data until we have an insufficent number of bytes to continue...
        while(cursor.hasNext(channel.type))
        {
            // Parse data into channel's data array
            if (channel.dataArrayPtr) {
                var value = cursor.next(channel.type);
                channel.dataArrayPtr.push(value);
                //gc.console.log("pambCustomCodec.js", channel.name + " " + channel.type + " " + value);
                
            } else {
                
                // IMPORTANT: We are assuming that only the frameID channel will have a null 'data' object!!!
                this.frameId = cursor.next(channel.type);
                if (this.frameId >= dataFrames.length) {
                    gc.console.error("pambCustomCodec.js", "Data parsing misalignment: Invalid frame id: " + this.frameId + "!");
                }
                currentFrame = dataFrames[this.frameId];
                //gc.console.log("pambCustomCodec.js", "frame ID: " + this.frameId);
            }
            this.channelCounter = (this.channelCounter + 1) % currentFrame.length;
            channel = currentFrame[this.channelCounter];
        }
        
        // Keep the residual data, remove the parsed data
        // NOTE: http://www.tothenew.com/blog/javascript-splice-vs-slice/
        binaryDataArray.splice(0, cursor.index);    // TODO: Test performance against Array.shift()
        
        
        // TODO: Use inline array function for better performance
        // reference: https://gamealchemist.wordpress.com/2013/05/01/lets-get-those-javascript-arrays-to-work-fast/
    };
    
    // Call this method before beginning a new data collections
    BinaryDataParser.prototype.clear = function(dataFrames) {
        
        // Reset internal variaables
        this.frameId = 0;
        this.channelCounter = 0;
        
        // Clear all channel data arrays
        dataFrames.forEach((frame) => {
            frame.forEach((channel) => {
                if (channel.dataArrayPtr) {
                    channel.dataArrayPtr.length = 0;   // Clear array by setting it's length to zero
                }
            });
        });
    };

    return BinaryDataParser;
    
})();

// Channel class used by BinaryDataParser and FrameDefinitions for sorting data into various "channels" 
var CreateChannel = function(nameStr, byteLen, isSigned, dataArray) {
    this.name = nameStr;
    this.size = byteLen;
    this.type = (isSigned ? "Int" : "Uint") + (8*byteLen);  // NOTE: This currently does not support float types!
    this.dataArrayPtr = dataArray;
};

// Create an instance of the BinaryDataParser class (TODO: Move this somewhere else)
var Parser = new BinaryDataParser();


//
// Binary data frame definitions
//

// Channel parameters
var channelCount = 4;   // Default value; may change at runtime 
var bytesPerWord = 3;   // Default value; may change at runtime 

// Data placeholders...  // TODO: Where to attach these arrays, model or codec?
var statusArray = [];
var channelData = [];
var crcArray = [];

var FrameDefinitions = [];  // NOTE: This array may need to be dynamically built if the data formatting is configurable by the user!

// FRAME 0
FrameDefinitions[0x00] = [];    // NOTE: When building a 2D array, we need to initialize each frame separately.

    // Frame ID (required) 
    FrameDefinitions[0x00].push(new CreateChannel("frameId", 1, false, null));
    
    // STATUS
    FrameDefinitions[0x00].push(new CreateChannel("status", 2, false, statusArray));
    
    // CHANNELS 
    for (let i = 0 ; i < channelCount; i++) {
        // NOTE: We create a channel data placeholder for all channels, even if they are disabled
        channelData[i] = [];
        FrameDefinitions[0x00].push(new CreateChannel("channel" + i, bytesPerWord, true, channelData[i]));
    }
    
    // CRC
    FrameDefinitions[0x00].push(new CreateChannel("crc", 2, false, crcArray));
    

